#ifndef INCLUDE_OSCILLATOR_H
#define INCLUDE_OSCILLATOR_H
/*
 * This is the Loris C++ Class Library, implementing analysis, 
 * manipulation, and synthesis of digitized sounds using the Reassigned 
 * Bandwidth-Enhanced Additive Sound Model.
 *
 * Loris is Copyright (c) 1999-2004 by Kelly Fitz and Lippold Haken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Oscillator.h
 *
 * Definition of class Loris::Oscillator, a Bandwidth-Enhanced Oscillator.
 *
 * Kelly Fitz, 31 Aug 1999
 * loris@cerlsoundgroup.org
 *
 * http://www.cerlsoundgroup.org/Loris/
 *
 */

#include "NoiseGenerator.h"
#include <vector>

//	begin namespace
namespace Loris {

class Breakpoint;

// ---------------------------------------------------------------------------
//	class Oscillator
//
//	Class Oscillator represents the state of a single bandwidth-enhanced
//	sinusoidal oscillator used for synthesizing sounds from Reassigned
//	Bandwidth-Enhanced analysis data. Oscillator encapsulates the oscillator
//	state, including the instantaneous radian frequency (radians per
//	sample), amplitude, bandwidth coefficient, and phase, and a 
//	bandlimited stochastic modulator. 
//
//	Class Synthesizer uses an instance of Oscillator to synthesize
//	bandwidth-enhanced Partials.
//
class Oscillator
{
//	--- implementation _--

	//	stochastic modulators:
	NoiseGenerator bwModulator;
	
	//	instantaneous oscillator state:
	double i_frequency;	//	radians per sample
	double i_amplitude;	//	absolute
	double i_bandwidth;	//	bandwidth coefficient (noise energy / total energy)
	
	//	accumulating phase state:
	double determ_phase;	//	deterministic phase in radians

//	--- interface ---
public:
//	--- construction --_
	Oscillator( void );
	/*	Construct a new Oscillator with all state parameters initialized
		to 0.
	 */
	 
	 
	//	Copy, assignment, and destruction are free.
	//
	// 	Copied and assigned Oscillators have the duplicate state
	//	variables and the filters have the same coefficients,
	//	but the state of the filter delay lines is not copied.

//	-- the new way --
	 
	void resetEnvelopes( const Breakpoint & bp, double srate );
	/*	Reset the instantaneous envelope parameters 
	 	(frequency, amplitude, bandwidth, and phase).
	 	The sample rate is needed to convert the 
	 	Breakpoint frequency (Hz) to radians per sample.
	  */
	  
	void resetPhase( double ph );
	/*	Reset the phase of the Oscillator to the specified
		value, and clear the accumulated phase modulation. (?)
		Or not.
		This is done when the amplitude of a Partial goes to 
		zero, so that onsets are preserved in distilled
		and collated Partials.
	 */

	void oscillate( double * begin, double * end,
					const Breakpoint & bp, double srate );
	/*	Accumulate bandwidth-enhanced sinusoidal samples modulating the
		oscillator state from its current values of radian frequency, amplitude,
		and bandwidth to the specified target values. Accumulate samples into
		the half-open (STL-style) range of doubles, starting at begin, and
		ending before end (no sample is accumulated at end). The caller must
		insure that the indices are valid. Target frequency and bandwidth are
		checked to prevent aliasing and bogus bandwidth enhancement.
	 */

	/*	Access parameters of the Oscillator
		(why were these removed?)
	 */
	double amplitude( void ) const { return i_amplitude; }
	double bandwidth( void ) const { return i_bandwidth; }
	double phase( void ) const { return determ_phase; }
	double radianFreq( void ) const { return i_frequency; }

	/*	Used to allow state mutation too, why not anymore? */
	 
};	//	end of class Oscillator

}	//	end of namespace Loris

#endif /* ndef INCLUDE_OSCILLATOR_H */
